/*
 *	chasend.c
 *
 *	last modified by M.Izumo <masana-i@is.aist-nara.ac.jp>, Tue Feb 11 1997
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <errno.h>
#include <sys/wait.h>

extern int errno;

#include "chasen.h"
#include "pat.h"

#ifdef KOCHA
#define CHA_NAME       "KoCha"
#define CHA_PROG       "kocha"
#else
#define CHA_NAME       "ChaSen"
#define CHA_PROG       "chasen"
#endif
#define VERSION        " 1.0 protocol 1.0 chasend 1.6.1" /* Must be contained EUC code */
#ifndef CHACEND_PORT
#define CHACEND_PORT 31000	/* 㥻, äȶ줷(^^; */
#endif /* CHACEND_PORT */
#define DATA_BUFFER_SIZE (1024*1024) /* 1M */
#ifndef SOMAXCONN
#define SOMAXCONN 5
#endif /* SOMAXCONN */

#ifdef sun
#define memmove(a, b, n) bcopy(b, a, n)
#endif /* sun */

extern char Dic_file_name[MAX_DIC_NUMBER][1024]; /* ե̾(init.c) */
static int obj_dic_no = 0; /* ưŪ(ɲ)оݤȤʤ뼭ֹ */
static char *chasenrc_path;
static int client_fd = -1;
static FILE* client_ofp = NULL;
static FILE* client_ifp = NULL;
static char* data_buff = NULL;
static int data_column = 0;
static int data_buff_cur = 0;
static char* chacen_default_format = NULL;
static char chasen_out_format[BUFSIZ];
int chasend_error_flag = 0;
int skip_until_dot(FILE* fp);
int read_until_dot(char* buff, int bufsiz, FILE* fp);

void chasend_error(const char* fmt, ...)
{
    va_list args;

    if(chasend_error_flag)
	return;
    chasend_error_flag = 1;
    va_start(args, fmt);
    vfprintf(client_ofp, fmt, args);
    va_end(args);
}

void chasend_putc(int c)
{
    if(data_buff_cur < DATA_BUFFER_SIZE-10)
    {
	if(c != '\n')
	{
	    data_buff[data_buff_cur++] = c;
	    if(c == '.' && data_column == 0)
	    {
		data_buff[data_buff_cur++] = '.'; /* byte stuffing */
		data_column++;
	    }
	    data_column++;
	}
    }
}

void chasen_terminate_line(void)
{
    if(data_buff_cur < DATA_BUFFER_SIZE-10 && data_column)
    {
	data_buff[data_buff_cur++] = '\n';
	data_column = 0;
    }
}

void chasend_puts(const char* s)
{
    for(; *s; s++)
	chasend_putc(*s);
}

static void flush_buff(FILE* fp)
{
    if(!chasend_error_flag)
    {
	if(data_buff_cur > 0 && fp != NULL)
	{
	    fwrite(data_buff, 1, data_buff_cur, fp);
	    if(data_buff[data_buff_cur - 1] != '\n')
		putc('\n', fp);
	}

	if(fp != NULL)
	{
	    putc('.', fp);
	    putc('\n', fp);
	}
    }
    if(fp != NULL)
	fflush(fp);
    chasend_error_flag = 0;
    data_buff_cur = 0;
    data_column = 0;
}

void chasend_printf(const char* fmt, ...)
{
    va_list args;
    static char* tmpbuff = NULL;
    char* p;

    if(tmpbuff == NULL)
    {
	tmpbuff = (char *)malloc(CHA_INPUT_SIZE * 2);
	data_buff = (char *)malloc(DATA_BUFFER_SIZE);
	if(tmpbuff == NULL || data_buff == NULL)
	{
	    exit(1);
	}
	flush_buff(NULL);
    }

    va_start(args, fmt);
    if(fmt == NULL)		/* only initializing */
    {
	va_end(args);
	return;
    }

    vsprintf(tmpbuff, fmt, args);
    va_end(args);

    p = tmpbuff;
    while(*p)
	chasend_putc(*p++);
}



/*
 *  version()
 */
static void version(fp)
    FILE *fp;
{
    fprintf(fp, "%s version %s (C) 1996-1997 Nara Insittute of Science and Technology\n",
	    CHA_NAME, VERSION);
#ifndef VGRAM
    fprintf(fp, "Grammar files are compatible with JUMAN 2.0\n");
#else
    fprintf(fp, "Grammar files are in ChaSen's new format\n");
#endif
}

/*
 *  usage()
 */
static void usage()
{
    fprintf(stderr, "%s version %s (C) 1996-1997 Nara Insittute of Science and Technology\n",
	    CHA_NAME, VERSION);
    fprintf(stderr, "usage: %s [-h] [-V] [-r rc_file]\n",
	    CHA_PROG);
    fprintf(stderr, "  -r rc_file  use rc_file as a %src file other than the default\n",
	    CHA_PROG);
    fprintf(stderr, "  -h          print this help\n");
    fprintf(stderr, "  -V          print %s version\n", CHA_NAME);

    exit(1);
}

static int opt_show, opt_form, opt_ja, opt_cmd;

static void set_opt_form(format)
    char *format;
{
    char *f;
    f = format + strlen(format);
    if (f[-2] == '\\' && f[-1] == 'n')
      opt_form = 'F';
    else
      opt_form = 'W';
}

static char* chomp(char* s)
{
    int len = strlen(s);

    if(len < 2)
    {
	if(len == 0)
	    return s;
	if(s[0] == '\n' || s[0] == '\r')
	    s[0] = '\0';
	return s;
    }
    if(s[len - 1] == '\n')
	s[--len] = '\0';
    if(s[len - 1] == '\r')
	s[--len] = '\0';
    return s;
}


/*
------------------------------------------------------------------------------
  do_chasen()
------------------------------------------------------------------------------
*/
static int do_chasen(format)
    char *format;
{
    char line[CHA_INPUT_SIZE];
    char *(*get_func)();
    void (*print_func)();
    int len, istty;
    int i;
    char rslt[5000]; /* ѿ for ñå(exact match) */
    FILE *of; /* intե˽񤭹(ɲ)Τ */
    long new_word_index;
    char tmpstr[2000];

    /* jfgets.c */
    extern char *jfgets();
    extern char *fget_line();

    print_func = NULL;
    switch (opt_show) {
      case 'b': print_func = print_best_path; break;
      case 'm': print_func = print_all_mrph; break;
      case 'p': print_func = print_all_path; break;
    }

    /* ϤɸϤɤ */
    istty = isatty(fileno(stdout));

    while (fgets(line, sizeof(line), client_ifp) != NULL) {
        len = strlen(line);
	if(len >= sizeof(line) - 1)
	{
	    /* Too long input */
	    fputs("500 Line too long\n", client_ofp);
	    fflush(client_ofp);
	    flush_buff(NULL); /* throw the output buffer */
	    skip_until_dot(client_ifp);
	    return 1;
	}

	chomp(line);
	if(line[0] == '.' && line[1] == '\0')
	{
	    if(chasend_error_flag)
	    {
		flush_buff(client_ofp);
		return 1;
	    }
	    fputs("200 OK\n", client_ofp);
	    flush_buff(client_ofp);
	    fflush(client_ofp);
	    return 1;		/* Success */
	}

        /* remove stuff byte '.' */
	/* s/^\.\././; */
	if(line[0] == '.' && line[1] == '.')
	{
	    len = strlen(line);
	    memmove(line, line + 1, len); /* move memory including '\0' */
	}

	if (chasen_sent(line) == FALSE)
	  continue;

	if(chasend_error_flag)
	{
	    skip_until_dot(client_ifp);
	    flush_buff(client_ofp);
	    return 1;
	}
	else
	{
	    print_func(opt_form, format);
	    if(chasend_error_flag)
	    {
		skip_until_dot(client_ifp);
		flush_buff(client_ofp);
		return 1;
	    }

	    if (opt_form == 'W')
/*		chasend_putc('\n') */;
	    else if (opt_form != 'd')
	    {
		chasend_puts("EOS");
		chasen_terminate_line();
	    }
	}
    }

    if(!chasend_error_flag)
	fputs("200 OK\n", client_ofp);
    flush_buff(client_ofp);
    return 0;			/* Connection Closed */
}

/*
------------------------------------------------------------------------------
  chasen_run()
------------------------------------------------------------------------------
*/
static int chasen_run(int argc, char **argv)
{
    char *format = NULL;
    char *op;

    opt_show = 'b';
    opt_form = opt_cmd = 0;

    set_progname(CHA_PROG);
#ifdef _WIN32
    set_progpath(argv[0]);
#endif

    for (argv++; *argv && argv[0][0] == '-' && argv[0][1]; argv++) {
	for (op = &argv[0][1]; *op; op++) {
	    switch (*op) {
	      case 'b':
	      case 'm':
	      case 'p':	opt_show = *op;	break;
	      case 'f':
	      case 'e':
	      case 'c':
	      case 'd': opt_form = *op; break;
	      case 'F':
		if (!*++argv)
		{
		    fprintf(client_ofp,
			    "501 option requires an argument -- %c\n", *op);
		    fflush(client_ofp);
		    return 1;
		}
		if(format != NULL && format != chacen_default_format)
		    free(format);
		format = cha_strdup(*argv);
		break;
	      case 'C': opt_cmd = 1; break;
	      default:
		fprintf(client_ofp, "501 invalid option -- %c\n", *op);
		fflush(client_ofp);
		return 1;
	    }
	}
    }

    if(format == NULL)
	format = chacen_default_format;

    /*
     *  (by O.I.):
     *  -d ϼưŪ -m ޤ
     */
    if (opt_form == 'd')
      opt_show = 'm';

    if (!opt_form) {
	if (!format)
	  opt_form = 'f';
	else
	  set_opt_form(format);
    }

    /* opt_show  'b' ʤ饳 0 ѹ */
    if (opt_show == 'b')
      Cost_width = 0;

    return do_chasen(format);
}

static void expand_string(char* in)
{
    int i, n;
    char* out;

    n = strlen(in);
    out = in;

    for(i = 0; i < n; i++)
    {
	if(in[i] != '\\')
	    *out++ = in[i];
	else
	{
	    i++;
	    switch(in[i])
	    {
	      case 'n': *out++ = '\n'; break;
	      case 't': *out++ = '\t'; break;
	      case 'v': *out++ = '\v'; break;
	      case 'b': *out++ = '\b'; break;
	      case 'r': *out++ = '\r'; break;
	      case 'f': *out++ = '\f'; break;
	      case 'a': *out++ = 0x07; break;
	      case '\0': break;
	      default:
		*out++ = in[i];
	    }
	}
    }
    *out = '\0';
}

static int split_args(char* argbuff, int maxargc, char** argv)
{
    char* argstart;
    char* argend;
    char* p;
    int argc, i;

    argstart = argbuff;
    maxargc--;

    for(argc = 0; argc < maxargc; argc++)
    {
	/* skip space */
	while(*argstart == ' ')
	    argstart++;

	if(*argstart == '\0')
	    break;
	argv[argc] = argstart;

	/* find end of arg. */
	argend = argstart;
	while(*argend != '\0' && *argend != ' ')
	{
	    if(*argend != '\\')
		argend++;
	    else
	    {
		argend++;
		if(*argstart != '\0')
		    argend++;
	    }
	}
	if(*argend == '\0')
	{
	    argc++;
	    break;
	}
	*argend = '\0';
	argstart = argend + 1;
    }
    argv[argc] = NULL;

    for(i = 0; i < argc; i++)
	expand_string(argv[i]);

    return argc;
}

/* return:
 * 1 - continue
 * 0 - connection closed
 */
static int do_cmd(char* line)
{
    char* argv[16];
    int argc;

    argc = split_args(line, 16, argv);

/*    {int i;
    for(i = 0; i < argc; i++)
	printf("##[%d] <%s>\n", i, argv[i]);} */

    if(!strcasecmp(argv[0], "RUN"))
    {
	return chasen_run(argc, argv);
    }
    if(!strcasecmp(argv[0], "QUIT"))
    {
	return 0;
    }
    if(!strcasecmp(argv[0], "HELP"))
    {
	fprintf(client_ofp, "200 OK\n");
	fprintf(client_ofp,
	    "RUN [-bmp] [-fecd] [-F format]\n"
	    "  -b          show best path (default)\n"
	    "  -m          show all morphemes\n"
	    "  -p          show all paths\n"
	    "\n"
	    "  -f          show formatted morpheme data (default)\n"
	    "  -e          show entire morpheme data\n"
	    "  -c          show coded morpheme data\n"
	    "  -d          show detailed morpheme data\n"
	    "  -F format   show morpheme with formatted output\n"
	    );
	fprintf(client_ofp, "\n");
	fprintf(client_ofp, "QUIT\n");
	fprintf(client_ofp, "HELP\n");
	fprintf(client_ofp, ".\n");
	fflush(client_ofp);
	return 1;
    }
    else
    {
	fprintf(client_ofp, "500 What ?\n");
	fflush(client_ofp);
	return 1;
    }
    return 1;			/* dmy */
}

static void cmd_loop(void)
{
    static char buff[BUFSIZ];
    char* p;

    while(fgets(buff, BUFSIZ, client_ifp))
    {
	chomp(buff);
	if(!buff[0])
	    continue;
	if(!do_cmd(buff))
	    break;
    }

    fprintf(client_ofp, "205 Quit\n");
    fflush(client_ofp);
}

static int open_server_socket(unsigned short port)
{
    int sfd;
    struct sockaddr_in server;

    if((sfd = socket(AF_INET, SOCK_STREAM, NULL)) < 0)
    {
	perror("sokcet");
	return -1;
    }

    memset(&server, 0, sizeof(server));
    server.sin_port        = htons(port);
    server.sin_family      = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY);

#ifdef SO_REUSEADDR
    {
	int on = 1;
	setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (caddr_t)&on, sizeof(on));
    }
#endif /* SO_REUSEADDR */

    if(bind(sfd, (struct sockaddr *)&server, sizeof(server)) < 0)
    {
	perror("bind");
	close(sfd);
	return -1;
    }

    /* Set it up to wait for connections. */
    if(listen(sfd, SOMAXCONN) < 0)
    {
	perror("listen");
	close(sfd);
	return -1;
    }

    return sfd;
}

void sigchld_handler(int sig)
{
    int status;

    while(wait3(&status, WNOHANG, NULL) > 0)
	;
    signal(SIGCHLD, sigchld_handler);
}

static int sfd = -1;
void sig_term()
{
    shutdown(client_fd, 2);
    shutdown(sfd, 2);
    exit(0);
}

void main(int argc, char** argv)
{
    char *format;
    char *op;
    char *p;
    int pid, i;
    extern void set_chasenrc_path(); /* iotool.c */

    chasend_printf(NULL);
    format = NULL;
    opt_show = 'b';
    opt_form = opt_cmd = 0;

    if((p = strrchr(argv[0], '/')) == NULL)
	set_progname(argv[0]);
    else
	set_progname(p + 1);

    for (argv++; *argv && argv[0][0] == '-' && argv[0][1]; argv++) {
	for (op = &argv[0][1]; *op; op++) {
	    switch (*op) {
	      case 'r':
		if (!*++argv)
		  cha_exit(1, "option requires an argument -- %c", *op);
		set_chasenrc_path(*argv);
		op[1] = 0;
		break;
	      case 'V': version(stderr); exit(0);
	      case 'h': usage();
	      default:
		fprintf(stderr, "%s: invalid option -- %c\n", argv[0], *op);
		usage();
	    }
	}
    }

    /* daemon initialize */
    umask(0);
    if((i = fork()) > 0)
	exit(0);
    else if(i == -1) {
	fprintf(stderr,"chasend: unable to fork new process\n");
	perror("fork");
	exit(1);
    }
    if(setsid() == -1)
	perror("Warning: setsid");
    signal(SIGHUP,  SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGTERM, sig_term);
    signal(SIGINT,  sig_term);
    signal(SIGQUIT, sig_term);

    /* chasen initialize */
    chasen_init(&chacen_default_format);

    /* make a socket */
    if((sfd = open_server_socket(CHACEND_PORT)) < 0)
	exit(1);

#if 0
    printf("## Ready to connection\n");
    printf("## Parent PID=%d\n", getpid());
#endif

    signal(SIGCHLD, sigchld_handler);

    while(1)
    {
	int pid;

	if((client_fd = accept(sfd, NULL, NULL)) < 0)
	{
	    if(errno == EINTR)
		continue;
	    perror("accept");
	    exit(1);
	}

	if((pid = fork()) < 0)
	{
	    perror("fork");
	    sleep(1);
	    continue;
	}

	if(pid == 0)		/* child */
	{
	    close(sfd);

	    client_ofp = fdopen(client_fd, "w");
	    client_ifp = fdopen(client_fd, "r");

	    fprintf(client_ofp, "200 Running chasen version: " VERSION "\n");
	    fflush(client_ofp);

	    cmd_loop();

	    shutdown(client_fd, 2);
	    fclose(client_ofp);
	    fclose(client_ifp);
	    close(client_fd);
	    exit(0);
	}

	close(client_fd);
    }
}
