/*
 *	init.c
 *
 *	last modified by A.Kitauchi <akira-k@is.aist-nara.ac.jp> Sep. 1996
 *
 */

#include "chasen.h"
#include "pat.h"

#define DEF_UNDEF       "̤"
#define DEF_U_CLASS     "̤ʻ"
#define DEF_CLASS_C     "ʻ쥳"
#define DEF_RENSETSU_W  "ϢܥȽŤ"
#define DEF_KEITAISO_W  "ǥȽŤ"
#define DEF_COST_HABA   ""
#define DEF_U_CON_COST  "̤Ϣܥ"
#define DEF_COMPOSITION "Ϣʻ"
/*#define DEF_DIC_FILE    "ե"*/
#define DEF_DIC_FILE    "PATDIC" /* changed by Tatuo 960920 */

/* .chasenrc default values */
#define CLASS_COST_DEFAULT	10
#define RENSETSU_WEIGHT_DEFAULT	10
#define KEITAISO_WEIGHT_DEFAULT	1
#define COST_WIDTH_DEFAULT	20
#define UNDEF_WORD_DEFAULT	10000

char Dic_file_name[MAX_DIC_NUMBER][FILENAME_MAX]; /* ե̾ */

int Con_cost_weight, Mrph_cost_weight, Cost_width;
int Undef_cost, Undef_hinsi, Undef_con_tbl, Undef_con_cost;
#ifndef VGRAM
int Undef_bunrui;
#endif

/*
------------------------------------------------------------------------------
        PROCEDURE: <read_class_cost>
------------------------------------------------------------------------------
*/
static void read_class_cost(cell)
    cell_t *cell;
{
    cell_t *cell1;
    int hinsi, cost;

#ifdef VGRAM
    char *s;

    for (; !nullp(cell); cell = cdr(cell)) {
	cell1 = car(cell);
	cost = atoi(s_atom(car(cdr(cell1))));
	s = s_atom(car(car(cell1)));
	if (!strcmp(DEF_UNDEF, s))
	  Undef_cost = cost;
	else if (!strcmp(s, "*")) {
	    for (hinsi = 1; Hinsi[hinsi].name; hinsi++)
	      if (Hinsi[hinsi].cost == 0)
		Hinsi[hinsi].cost = cost;
	} else {
	    int match = 0;
	    for (hinsi = 1; Hinsi[hinsi].name; hinsi++) {
		if (match_nhinsi(car(cell1), hinsi)) {
		    Hinsi[hinsi].cost = cost;
		    match = 1;
		}
	    }
	    if (!match) {
		cha_exit_file(-1, "invalid hinsi name `");
		s_print(stderr, cell1);
		fputs("'\n", stderr); 
		exit(ConfigError);
	    }
	}
    }

    /* default */
    for (hinsi = 1; Hinsi[hinsi].name; hinsi++) 
      if (Hinsi[hinsi].cost == 0)
	Hinsi[hinsi].cost = CLASS_COST_DEFAULT;

    /* ʸƬ ʸ */
    Hinsi[0].cost = 0;
#else
    int bunrui;

    for (; !nullp(cell); cell = cdr(cell)) {
	cell1 = car(cell);
	cost = atoi(s_atom(car(cdr(cell1))));
	if (!strcmp(DEF_UNDEF, s_atom(car(car(cell1)))))
	  Undef_cost = cost;
	else {
	    get_hinsi_bunrui_id(car(cell1), &hinsi, &bunrui);
	    if (hinsi == 0) {
		for (hinsi = 1; Class[hinsi][0].id; hinsi++)
		  for (bunrui = 0; Class[hinsi][bunrui].id; bunrui++)
		    if (Class[hinsi][bunrui].cost == 0)
		      Class[hinsi][bunrui].cost = cost;
	    } else if (bunrui == 0) {
		for (bunrui = 0; Class[hinsi][bunrui].id; bunrui++)
		  Class[hinsi][bunrui].cost = cost;
	    } else {
		Class[hinsi][bunrui].cost = cost;
	    }
	}
    }

    /* default */
    for (hinsi = 1; Class[hinsi][0].id; hinsi++) 
      for (bunrui = 0; Class[hinsi][bunrui].id; bunrui++)
	if (Class[hinsi][bunrui].cost == 0)
	  Class[hinsi][bunrui].cost = CLASS_COST_DEFAULT;

    /* For ʸƬ ʸ added by S.Kurohashi */
    Class[0][0].cost = 0;
#endif
}

static void read_composition(cell)
    cell_t *cell;
{
#ifdef VGRAM
    int comp, pos;

    for (; !nullp(cell); cell = cdr(cell)) {
	comp = get_nhinsi_id(car(cell));
	for (pos = 1; Hinsi[pos].name; pos++)
	  if (match_nhinsi(car(cell), pos))
	    Hinsi[pos].comp = comp;
    }
#else
    int comp_hinsi, comp_bunrui, i;

    for (; !nullp(cell); cell = cdr(cell)) {
	get_hinsi_bunrui_id(car(cell), &comp_hinsi, &comp_bunrui);
	if (comp_bunrui)
	  Class[comp_hinsi][comp_bunrui].comp = 2;
	else {
	    for (i = 0; Class[comp_hinsi][i].id; i++)
	      Class[comp_hinsi][i].comp = 1;
	}
    }
#endif		
}

/**************** ѥڥǡɤ by TATUO ***************/
void read_patdic(cell)
    cell_t *cell;
{
    int  num;
    char tmpstr[FILENAME_MAX];
    char patfile[FILENAME_MAX];

    num = 0;
    for (; !nullp(cell); cell = cdr(cell)) {
	if (num >= MAX_DIC_NUMBER)
	  cha_exit_file(ConfigError, "too many dictionary files");
	strcpy(Dic_file_name[num], s_atom(car(cell)));

	/* Υץ */
	strcpy(tmpstr,Dic_file_name[num]);
	strcat(tmpstr,".int");

	/* .chasenrc ɤ߹ʤΤ cha_fopen ϻȤʤ */
	if ((dic_file[num] = fopen(tmpstr, "r")) == NULL)
	  cha_exit(1, "can't open %s", tmpstr);

	strcpy(patfile,Dic_file_name[num]);
	strcat(patfile, ".pat");

	(void)pat_init_tree_top(&tree_top[num]);
	(void)com_l(patfile,&tree_top[num]);
	num++;
    }
    number_of_tree = num;
}

/*
------------------------------------------------------------------------------
	PROCEDURE: <chasen_init>              >>> changed by T.Nakamura <<<
------------------------------------------------------------------------------
*/
char *chasen_init(formatp)
    char **formatp;
{
    FILE *fp;
    cell_t *cell1, *cell2;
    char *rc_path, *cell1_str;

    rc_path = read_chasenrc();   /* by T.U */
    read_grammar(NULL, 1, 1);

    /* DEFAULT VALUE*/
    Mrph_cost_weight = KEITAISO_WEIGHT_DEFAULT;
    /* modified by S.Kurohashi 93/02/18 
       for balancing connect-weigh with morph-weight */
    Con_cost_weight = RENSETSU_WEIGHT_DEFAULT * MRPH_DEFAULT_WEIGHT;
    Cost_width = COST_WIDTH_DEFAULT * MRPH_DEFAULT_WEIGHT;
    Undef_cost = UNDEF_WORD_DEFAULT;  
    Undef_con_cost = 0;

    fp = cha_fopen(rc_path, "r", 1);

    while (!s_feof(fp)) {
	cell1 = s_read(fp);
	cell1_str = s_atom(car(cell1));
	cell2 = car(cdr(cell1));

	/* ե */
	if (!strcmp(DEF_DIC_FILE, cell1_str))
	  read_patdic(cdr(cell1));
	/* ̤ʻ */
	else if (!strcmp(DEF_U_CLASS, cell1_str)) {
#ifdef VGRAM
	    Undef_hinsi = get_nhinsi_id(cell2);
#else
	    get_hinsi_bunrui_id(cell2, &Undef_hinsi, &Undef_bunrui);
#endif
	}
	/* ϢܥȽŤ*/	  
	/* modified by S.Kurohashi 93/03/04 
	   for balancing connect-weigh with morph-weight */
	else if (!strcmp(DEF_RENSETSU_W, cell1_str))
	  Con_cost_weight = atoi(s_atom(cell2)) * MRPH_DEFAULT_WEIGHT;
	/* ǥȽŤ*/
	else if (!strcmp(DEF_KEITAISO_W, cell1_str))
	  Mrph_cost_weight = atoi(s_atom(cell2));
	/*  */
	/* modified by S.Kurohashi 93/03/04 
	   for balancing connect-weigh with morph-weight */
	else if (!strcmp(DEF_COST_HABA, cell1_str))
	  Cost_width = atoi(s_atom(cell2)) * MRPH_DEFAULT_WEIGHT;
	/* ʻ쥳*/
	else if (!strcmp(DEF_CLASS_C, cell1_str))
	  read_class_cost(cdr(cell1));
	/* ̤Ϣܥ */
	else if (!strcmp(DEF_U_CON_COST, cell1_str))
	  Undef_con_cost = (int) atoi(s_atom(cell2));
	/* Ϣʻ */
	else if (!strcmp(DEF_COMPOSITION, cell1_str))
	  read_composition(cdr(cell1));
	/* ϥեޥå */
	else if (!*formatp && !strcmp("ϥեޥå", cell1_str)) {
	    char *s1, *s2;
	    s1 = s_atom(cell2);
	    s2 = s1 + strlen(s1) - 1;
	    if (*s1 == '"' && *s2 == '"') {
		s1++;
		*s2 = '\0';
	    }
	    *formatp = cha_strdup(s1);
	}
    }

    fclose(fp);

    read_katuyou(NULL, 1);
    read_table(NULL, 1);
    read_matrix(NULL);
#ifdef VGRAM
    Undef_con_tbl = check_table_for_undef(Undef_hinsi);
#else
    Undef_con_tbl = check_table_for_undef(Undef_hinsi, Undef_bunrui);
#endif

    return rc_path;
}

