/*
 * jfgets.c - fgets() for japaneses
 * 		by k-chinen@is.aist-nara.ac.jp, 1996.
 *
 * $Id: jutils.c,v 1.1 1996/03/11 19:43:28 k-chinen Exp k-chinen $
 * 
 * NOTE: An idea of these procedures are taken from youhcan's jutils.c
 *	 for wais-japanese
 *
 */


#include <stdio.h>
#include <string.h>

#define DELIMIT2CHARS ""
#define DELIMIT1CHARS ",."



#define INNER_BUFSIZE   8192


/*
 * pretty printing for debug
 */

void
print_string(prefix, postfix, buf)
     unsigned char *prefix, *postfix, *buf;
{
  unsigned char tmp_buf[INNER_BUFSIZE];

  char *p, *q;

  p = (char *)buf;
  q = (char *)tmp_buf;
  while(*p) {
    if(*p=='\n') {
      *q++ = '\\';
      *q++ = 'n';
      p++;
    }
    else
    if(*p=='\t') {
      *q++ = '~';
      p++;
    }
    else
      *q++ = *p++;
  }
  *q = '\0';

  fprintf(stderr, "%s%s%s", prefix, tmp_buf, postfix);
  fflush(stderr);
}





static
int
jistoeuc(ibuffer, obuffer)
     unsigned char *ibuffer;
     unsigned char *obuffer;
{
  register unsigned char *p, h, l;
  register int level, flag, c;

  level = 0;
  flag = 0;

  p = ibuffer;
  c = 0;
  while(*p) {
    if(*p==0x1b) {
      level = 1;
    }
    else {
      if(level) {
          if(level==1) {
              if(*p=='$')         level = 2;
              else if(*p=='(')    level = 12;
          }
          else {
              /* Translation */
              if(level== 2 && *p=='@')    flag = 1;
              if(level== 2 && *p=='B')    flag = 1;
              if(level==12 && *p=='B')    flag = 0;
              if(level==12 && *p=='J')    flag = 0;

              /* Give up to parse escape sequence */
              level=0;
          }
      }
      else {
          /* KANJI mode without control charactors */
          if(flag && *p>=0x20) {
              h = *p;
              p++;
              l = *p;

              obuffer[c++] = h | 0x80;
              obuffer[c++] = l | 0x80;
          }
          /* ASCII mode or control charactor in KANJI mode */
          else {
              obuffer[c++] = *p;
          }
      }
    }
    p++;
  }
  obuffer[c] = '\0';

  return 0;
}





/*
 * reduce_white - reduce whites in string with this rules.
 *
 *      [KANJI WHITES KANJI] -> [KANJI1 KANJI2]
 *      [KANJI WHITES ASCII] -> [KANJI SPACE ASCII]
 *      [ASCII WHITES ASCII] -> [ASCII SPACE ASCII]
 *      [ASCII WHITES KANJI] -> [ASCII SPACE KANJI]
 */

void
reduce_white(orig)
     unsigned char *orig;
{
  unsigned char *base, *last;
  unsigned char *first, *cur;

#if 0
  fprintf(stderr, "*before '%s'\n", orig);
#endif

  last = NULL;
  base = orig;
  while(*base) {
    if(*base==' '||*base=='\t') {
      cur = base;
      cur++;
      while(*cur==' '||*cur=='\t') {
        cur++;
      }

      first = base;
      if(last!=NULL) {
        if(*cur>= 0xa0 && *last >= 0xa0) {
        }
        else {
          *first = ' ';
          first++;
          /*
          base++;
          */
        }
      }
#if 0
      fprintf(stderr, " change base %4d first %4d cur %4d last '%c'\n",
                base-orig, first-orig, cur-orig, last==NULL ? 'X' : *last );
#endif
      memcpy(first, cur, strlen(cur)+1);
#if 0
      fprintf(stderr, "*change '%s'\n", orig);
#endif
    }

    last = base;
    base++;
  }

#if 0
  fprintf(stderr, "*after  '%s'\n", orig);
  fflush(stderr);
#endif
}


/*
 * isterminator - check it is terminator or not
 *
 * return
 *		 1: terminator
 *		 0: not terminator
 *		-1:	error
 */

int
isterminator(target, termlist)
     unsigned char *target;
     unsigned char *termlist;
{
	if(termlist==NULL || target==NULL) {
		return -1;
	}
	while(*termlist) {
		if(*termlist==*target) {
			if(*termlist & 0x80) {
				if(*(termlist+1) && *(target+1)
					&& *(termlist+1)==*(target+1)) {
					return 1;
				}
			}
			else {
				return 1;
			}
		}

		/* try next terminator */
		if(*termlist & 0x80) {
			termlist+=2;
		}
		else {
			termlist++;
		}
	}

	return 0;
}





/*
 * inner buffer and inner position.
 *      if stream is empty. 'pos' point NULL.
 *
 */

static unsigned char ibuf[INNER_BUFSIZE];
static unsigned char *pos=NULL;


/*
 * fget_line - get line via fgets(). So it is really reading function :-)
 *
 */

char *fget_line(buffer, bufsize, stream)
     char *buffer;
     int bufsize;
     FILE *stream;
{
  static unsigned char tmp_buf[INNER_BUFSIZE];
  char *chk;

  pos = (unsigned char *)buffer;

  if(feof(stream)) {
    buffer[0] = '\0';
    pos = NULL;

	return NULL;
  }

  if((chk=fgets(tmp_buf, bufsize, stream))==NULL) {
    buffer[0] = '\0';
    pos = NULL;

    return NULL;
  }

  /*
   * add '\n' at the end of the line
   *     by A.Kitauchi Nov. 1996
   */
  {
    char *s = (char *)tmp_buf + strlen(tmp_buf);
    if (s[-1] != '\n') {
      s[0] = '\n';
      s[1] = '\0';
    }
  }

  /*
   * call convertor
   * NOTE: EUC string is short than JIS string.
   *       if you want to other conversion. you must care string length.
   */

#if 0
  print_string("\n0<", ">\n",   tmp_buf);
#endif

  jistoeuc(tmp_buf, buffer);

#if 0
  print_string(  "1<", ">\n", buffer);
#endif

  return chk;
}



/*
 * jfgets - fgets() for Japanese Text.
 *
 */

char *
jfgets(buffer, bufsize, stream)
     char *buffer;
     int bufsize;
     FILE *stream;
{
  unsigned char *q, *cur;
  unsigned char k1, k2;
  int count;
  int kflag;            /* kanji flag(0=not found, 1=found) */


#if 0
  fprintf(stderr, "!!! %s\n", pos==NULL ? "EOF": "more");
#endif

  if(pos==NULL) {
    if(fget_line(ibuf, sizeof(ibuf), stream)==NULL) {
      return NULL;
    }
  }


  kflag = 0;
  q = (unsigned char *)buffer;
  count = bufsize;
  while(*pos && count>0) {
    k1 = *pos;
    count--;

    /* KANJI */
    if(*pos>=0xa0 && *(pos+1)) {
      kflag = 1;
	  cur = pos;

      k2 = *(pos+1);
      count--;
      pos+=2;

#if 0
      fprintf(stderr, "@@");
      fflush(stderr);
#endif

      *q++ = k1;
      *q++ = k2;

      /* hit delimiter */
      if(isterminator(cur, "")) {
#if 0
        fprintf(stderr, "!");
        fflush(stderr);
#endif
        goto out;
      }
    }
    /* not KANJI */
    else {
      pos++;

      /* line is end */
      if(k1=='\n') {
        if(fget_line(ibuf, sizeof(ibuf), stream)==NULL)
			goto out;

        /* have kanji, connect next line */
        if(kflag && count>0) {
            /* double '\n' is paragraph end. so it is delimiter */
            if(*pos=='\n')
              goto out;

            if(*pos>=0xa0) {
              /* thru */
            }
            else {
              *q++ = ' ';
            }
        }
        /* not have kanji or no space, return with this line */
        else {
          goto out;
        }
      }
      else {
#if 0
        fprintf(stderr, "_");
        fflush(stderr);
#endif

        *q++ = k1;
      }
    }

#if 1
    /* line is end without '\n', long string read more */
    if(*pos=='\0' && count>0) {
        fget_line(ibuf, sizeof(ibuf), stream);
    }
#endif
  }

out:
  *q = '\0';

#if 0
  fprintf(stderr, "\n1[%s]\n", buffer);
#endif
  reduce_white(buffer);
#if 0
  fprintf(stderr, "\n2[%s]\n", buffer);
#endif

  return buffer;
}

