/*
==============================================================================
	iotool.c
		1990/12/14/Fri	Yutaka MYOKI(Nagao Lab., KUEE)
		1990/12/25/Tue	Last Modified
		Oct. 1996       A.Kitauchi <akira-k@is.aist-nara.ac.jp>
==============================================================================
*/

#include "chadic.h"

#define DEF_GRAM_FILE "ʸˡե"

#ifdef KOCHA
#define RCFILE "/.kocharc"
#else
#ifdef _WIN32
#define RCFILE "\\chasenrc"
#else
#define RCFILE "/.chasenrc"
#endif
#endif
#define OLDRCFILE "/.jumanrc"

static char progname[FILENAME_MAX];
static char filepath[FILENAME_MAX];
static char grammar_dir[FILENAME_MAX];
static char chasenrc_path[FILENAME_MAX];
#ifdef _WIN32
static char progpath[FILENAME_MAX];
#endif
int LineNo, LineNoForError;

/***********************************************************************
 * set_progname - set program name
 *
 *	progname is used to in cha_exit() and cha_exit_file()
 ***********************************************************************/
void set_progname(name)
    char *name;
{
    strcpy(progname, name);
}    

#ifdef _WIN32
/***********************************************************************
 * set_progpath - set program pathname
 ***********************************************************************/
void set_progpath(path)
    char *path;
{
    strcpy(progpath, path);
}
#endif

/***********************************************************************
 * set_chasenrc_path - set chasenrc file path
 *
 *	this function is called when -r option is used.
 ***********************************************************************/
void set_chasenrc_path(filename)
    char *filename;
{
    strcpy(chasenrc_path, filename);
}    

/***********************************************************************
 * get_grammar_dir
 *
 *	called only from chasen.c
 ***********************************************************************/
char *get_grammar_dir()
{
    return grammar_dir;
}

/***********************************************************************
 * cha_fopen - open file, or error end
 *
 * inputs:
 *	ret - exit code (don't exit if ret < 0)
 ***********************************************************************/
FILE *cha_fopen(filename, mode, ret)
    char *filename, *mode;
    int ret;
{
    FILE *fp;

    if ((fp = fopen(filename, mode)) != NULL) {
	if (filename != filepath)
	  strcpy(filepath, filename);
	if (*mode == 'r')
	  LineNo = LineNoForError = 0;
    } else if (ret >= 0)
      cha_exit(ret, "can't open %s", filename);

    return fp;
}

FILE *cha_fopen2(filename1, filename2, mode, ret)
    char *filename1, *filename2, *mode;
    int ret;
{
    FILE *fp;

    if ((fp = cha_fopen(filename1, mode, -1)) != NULL)
      return fp;

    if ((fp = cha_fopen(filename2, mode, -1)) != NULL)
      return fp;

    cha_exit(ret, "can't open %s or %s", filename1, filename2);
}

/***********************************************************************
 * cha_fopen_grammar - open file from current or grammar directory
 *
 * inputs:
 *	dir - 0: read from current directory
 *	      1: read from grammar directory
 *	      2: read from current directory or grammar directory
 *
 *	ret - return the code when fopen() fails
 *
 * outputs:
 *	filepathp - file path string
 ***********************************************************************/
FILE *cha_fopen_grammar(filename, mode, ret, dir, filepathp)
    char *filename, *mode, **filepathp;
    int ret, dir;
{
    FILE *fp;
    char *s;

    *filepathp = filename;
    switch (dir) {
      case 0:
	/* ȥǥ쥯ȥ꤫ɤ߹ */
	return cha_fopen(filename, mode, ret);
      case 2:
	/* ȥǥ쥯ȥ꤫ɤ߹ */
	if ((fp = cha_fopen(filename, mode, -1)) != NULL)
	  return fp;
	/* don't break */
      case 1:
	/* ʸˡǥ쥯ȥ꤫ɤ߹ */
	/* ʸˡǥ쥯ȥ꤬ꤵƤʤ .chasenrc ɤ߹ */
	if (grammar_dir[0] == '\0')
	  read_chasenrc();
	strcpy(filepath, grammar_dir);
	/* filepath  '/' (MSDOS  '\\') ղ */
	s = filepath + strlen(filepath);
	if (s[-1] != PATH_DELIMITER) {
	    s[0] = PATH_DELIMITER;
	    s[1] = '\0';
	}
	strcat(filepath, filename);
	*filepathp = filepath;
	return cha_fopen(filepath, mode, ret);
    }
}

FILE *cha_fopen_grammar2(filename1, filename2, mode, ret, dir, filepathp)
    char *filename1, *filename2, *mode, **filepathp;
    int ret, dir;
{
    FILE *fp;

    if ((fp = cha_fopen_grammar(filename1, mode, -1, dir, filepathp)) != NULL)
      return fp;

    if ((fp = cha_fopen_grammar(filename2, mode, -1, dir, filepathp)) != NULL)
      return fp;

    cha_exit(ret, "can't open %s or %s", filename1, filename2);
}

#if 0
/*
------------------------------------------------------------------------------
        FUNCTION:
        <cha_feof>: if <fp> points to "EOF" return <TRUE> else return <FALSE>
------------------------------------------------------------------------------
*/

int cha_feof(fp)
    FILE *fp;
{
    int c;

    if ((c = fgetc(fp)) == EOF) {
	return TRUE;
    } else {
	ungetc(c, fp);
	return FALSE;
    }
}
#endif

/*
------------------------------------------------------------------------------
        PROCEDURE:
        <cha_malloc>: do "malloc" (library function) and cha_exit processing
------------------------------------------------------------------------------
*/

void *cha_malloc(n)
    int n;
{
    void *p;

    if ((p = (void *)malloc(n)) == NULL)
      cha_exit(AllocateError, "not enough memory. can't allocate.");

    return p;
}

#define CHA_MALLOC_SIZE (1024 * 64)
static char *cha_malloc_char(size)
    int size;
{
    static int idx = CHA_MALLOC_SIZE;
    static char *ptr;

    if (idx + size >= CHA_MALLOC_SIZE) {
	ptr = (char *)cha_malloc(CHA_MALLOC_SIZE);
	idx = 0;
    }

    idx += size;
    return ptr + idx - size;
}

char *cha_strdup(str)
    char *str;
{
    char *newstr;

    newstr = cha_malloc_char(strlen(str) + 1);
    strcpy(newstr, str);

    return newstr;
}

/*
------------------------------------------------------------------------------
        PROCEDURE
        <cha_exit>: print error messages on stderr and exit
------------------------------------------------------------------------------
*/

void cha_exit(status, format, a, b, c, d, e, f, g, h)
    int status;
    char *format, *a, *b, *c, *d, *e, *f, *g, *h;
{
    fprintf(stderr, "%s: ", progname);

    fprintf(stderr, format, a, b, c, d, e, f, g, h);

    if (status >= 0) {
	fputc('\n', stderr); 
	exit(status);
    }
}

void cha_exit_file(status, format, a, b, c, d, e, f, g, h)
    int status;
    char *format, *a, *b, *c, *d, *e, *f, *g, *h;
{
    if (LineNo == 0)
      fprintf(stderr, "\n%s: ", progname);
    else if (LineNo == LineNoForError)
      fprintf(stderr, "\n%s: %s:%d: ", progname, filepath, LineNo);
    else
      fprintf(stderr, "\n%s: %s:%d-%d: ", progname, filepath,
	      LineNoForError, LineNo);

    fprintf(stderr, format, a, b, c, d, e, f, g, h);

    if (status >= 0) {
	fputc('\n', stderr); 
	exit(status);
    }
}

/*
------------------------------------------------------------------------------
	FUNCTION:
	<compare_end_str>: if <s1> = <...s2> or <s2> = <...s1> return TRUE
------------------------------------------------------------------------------
*/

int compare_end_str(s1, s2)
    char *s1, *s2;
{
    int l1, l2;

    l1 = strlen(s1);
    l2 = strlen(s2);

    if (l1 >= l2) {
	if (strcmp(s1 + l1 - l2, s2) == 0)
	  return TRUE;
	else
	  return FALSE;
    } else {
	if (strcmp(s2 + l2 - l1, s1) == 0)
	  return TRUE;
	else
	  return FALSE;
    }
}

/*
------------------------------------------------------------------------------
	PROCEDURE:
	<ls>: print <char *:p> file-information on <FILE *:fp>
------------------------------------------------------------------------------
*/

#if 0
void ls(fp, p, f)
    FILE *fp;
    char *p, char *f;
{
    char path[FILENAME_MAX];
    struct stat stbuf;

    strcpy(path, p);
    strcat(path, f);
    stat(path, &stbuf);

    fprintf(fp, "%8ld bytes: %s\n", stbuf.st_size, path);
}
#endif

static FILE *fopen_chasenrc(rc_path)
    char *rc_path;
{
    FILE *fp;
    char *home_dir, *rc_env, *getenv();

    /* -r option */
    if (chasenrc_path[0]) {
	strcpy(rc_path, chasenrc_path);
	return cha_fopen(rc_path, "r", OpenError);
    }

    /* environment variable */
    if ((rc_env = getenv("CHASENRC")) != NULL) {
	strcpy(rc_path, rc_env);
	return cha_fopen(rc_path, "r", OpenError);
    }

    /* home directory */
    if ((home_dir = getenv("HOME")) == NULL) {
#ifndef _WIN32
      cha_exit(ConfigError, "can't get environment variable HOME.");
#endif
    } else {
	/* .chasenrc */
	strcpy(rc_path, home_dir);
	strcat(rc_path, RCFILE);
	if ((fp = cha_fopen(rc_path, "r", -1)) != NULL)
          return fp;

#ifndef _WIN32
	/* .jumanrc */
	strcpy(rc_path, home_dir);
	strcat(rc_path, OLDRCFILE);
	if ((fp = cha_fopen(rc_path, "r", -1)) != NULL)
          return fp;
#endif
    }

#ifdef _WIN32
    strcpy(rc_path, progpath);
    strcpy(strrchr(rc_path, '\\'), RCFILE);
    if ((fp = cha_fopen(rc_path, "r", -1)) != NULL)
      return fp;
#endif

#ifdef RCPATH
    /* RCPATH in Makefile */
    strcpy(rc_path, RCPATH);
    if ((fp = cha_fopen(rc_path, "r", -1)) != NULL)
      return fp;
#ifdef _WIN32
    cha_exit(OpenError, "can't open chasenrc or %s", rc_path);
#else
    cha_exit(OpenError, "can't open .chasenrc, .jumanrc, or %s", rc_path);
#endif
#else /* !RCPATH */
#ifdef _WIN32
    cha_exit(OpenError, "can't open chasenrc");
#else
    cha_exit(OpenError, "can't open .chasenrc or .jumanrc");
#endif
#endif /* !RCPATH */
}

/***********************************************************************
 * read_chasenrc - read .chasenrc and set grammar directory
 ***********************************************************************/
char *read_chasenrc()
{
    static char rc_path[FILENAME_MAX];
    FILE *fp;
    cell_t *cell;

    fp = fopen_chasenrc(rc_path);

    while (!s_feof(fp)) {
	cell = s_read(fp);
	if (!strcmp(DEF_GRAM_FILE, s_atom(car(cell))))
	  strcpy(grammar_dir, s_atom(car(cdr(cell))));
    }

    fclose(fp);

    return rc_path;
}

