/*
 *	parse.c - parse a sentence
 *
 *	last modified by A.Kitauchi <akira-k@is.aist-nara.ac.jp>, Oct. 1996
 */

#include "chasen.h"
#include "pat.h"

#define MRPH_NUM	        8192
#define PATH1_NUM		256

#define HANKAKU            	0x80
#define PRIOD            	0xa1a5
#define CHOON            	0xa1bc
#define KIGOU            	0xa3b0
#define SUJI           	        0xa3c0
#define ALPH            	0xa4a0
#define HIRAGANA                0xa5a0
#define KATAKANA                0xa6a0
#define GR                      0xb0a0
#define KANJI                   0xffff

mrph2_t Mrph[MRPH_NUM];
path_t Path[PATH_NUM];
int Path_num;


/***********************************************************************
 * cha_malloc_char, cha_malloc_short, cha_buffer_free
 ***********************************************************************/
#define CHA_MALLOC_SIZE (1024 * 64)
static char *buffer_ptr[100];
static int  buffer_ptr_num = 0;
static int  buffer_idx = CHA_MALLOC_SIZE;

static void *cha_malloc_char(size)
    int size;
{
    if (buffer_idx + size >= CHA_MALLOC_SIZE) {
	if (buffer_ptr_num == 99)
	  cha_exit(1, "Can't allocate memory");
	buffer_ptr[buffer_ptr_num++] = cha_malloc(CHA_MALLOC_SIZE);
	buffer_idx = 0;
    }

    buffer_idx += size;
    return buffer_ptr[buffer_ptr_num - 1] + buffer_idx - size;
}

static void *cha_malloc_short(nelem)
    int nelem;
{
    int size = nelem * sizeof(short);

    if (buffer_idx & 1)
      buffer_idx++;

    if (buffer_idx + size >= CHA_MALLOC_SIZE) {
	if (buffer_ptr_num == 99)
	  cha_exit(1, "Can't allocate memory");
	buffer_ptr[buffer_ptr_num++] = cha_malloc(CHA_MALLOC_SIZE);
	buffer_idx = 0;
    }

    buffer_idx += size;
    return buffer_ptr[buffer_ptr_num - 1] + buffer_idx - size;
}

static void cha_buffer_free()
{
    while (buffer_ptr_num > 0)
      free(buffer_ptr[--buffer_ptr_num]);
    buffer_idx = CHA_MALLOC_SIZE;
}

/***********************************************************************
 * check_code()
 ***********************************************************************/
static int check_code(str)
    char *str;
{
    int	code;
    unsigned char *s = (unsigned char *)str;

    /*  nyuuryoku chuuni hankaku space wo yurusu, by. T.U. '96.01.10 */
#if 1
    if (*s == '\0' || *s == ' ')	return 0;
#else
    if (*s == '\0')			return 0;
#endif
    else if (*s < HANKAKU)		return HANKAKU;

    code = *s * 256 + *(s + 1);

    if ( code == PRIOD ) 			return PRIOD;
    else if ( code == CHOON ) 			return CHOON;
    else if ( code < KIGOU ) 			return KIGOU;
    /* else if ( code < SUJI ) 			return SUJI; */
    else if ( code < ALPH )			return ALPH;
    else if ( code < HIRAGANA ) 		return HIRAGANA;
    else if ( code < KATAKANA ) 		return KATAKANA;
    else if ( code < GR ) 			return GR;
    else return KANJI;
}

/***********************************************************************
 * undef_mrph_len - ̤ĹĴ٤
 *
 * Ҥ餬ʡ: 1ʸ
 * ʡ޻ʤ: Ϣ³ʸ
 ************************************************************************/
static int undef_mrph_len(target)
    char *target;
{
    int	code, next_code;
    int len = 0;

    code = check_code(target);

    if ( code == HIRAGANA )
      return 2;
    if ( code == KANJI )
      return 2;

    while(1) {
	if ( code == HANKAKU )
	  len++;
	else
	  len += 2;
	next_code = check_code(target + len);
	if ( next_code == code
	    || ( code == KATAKANA && next_code == CHOON )
	    || ( code == ALPH     && next_code == PRIOD ) ) continue;
	else break;
    }
    return len;
}

/***********************************************************************
 *  ̤Хåեɲ
 ***********************************************************************/
static int register_undef_mrph(target, mrph_idx, undef_len)
    char *target;
    int mrph_idx, undef_len;
{
#if 0
    int undef_len;
#endif
    mrph2_t *mrph = &Mrph[mrph_idx];

#if 0
    undef_len = undef_mrph_len(target);
#endif

    mrph->midasi = target;
    mrph->base_length = mrph->length = undef_len;

    mrph->yomi = (char *)cha_malloc_char(undef_len + 1);
    strncpy(mrph->yomi, target, undef_len);
    mrph->yomi[undef_len] = '\0';

    mrph->hinsi = Undef_hinsi;
#ifndef VGRAM
    mrph->bunrui = Undef_bunrui;
#endif
    mrph->con_tbl = Undef_con_tbl;
    mrph->ktype = 0;
    mrph->kform = 0;
    mrph->is_undef = 1; /* ̤ */
    mrph->weight = MRPH_DEFAULT_WEIGHT;
    mrph->info = "NIL"; /* ̣ʸ NIL Ȥ롥 */

    if (++mrph_idx == MRPH_NUM)
      return FALSE;

    return TRUE;
}

/*
------------------------------------------------------------------------------
        PROCEDURE: <take_data>                  >>> added by T.Nakamura <<<
------------------------------------------------------------------------------
*/
char *take_data(mrph, pbuf, target)
    mrph2_t *mrph;
    char *pbuf, *target;
{
    char *ret;
    char *p, *s;
    int  i;

#if 1
    p = pbuf;

    /* ФĹ */
    for (i = 0; *p++ != '\t'; i++);
    mrph->base_length = mrph->length = i;
    /* ʻʬ No. */
    for (i = 0; *p != ' '; i = i * 10 + *p++ - '0');
    mrph->hinsi = i; p++;
#ifndef VGRAM
    /* ʻʬ No. */
    for (i = 0; *p != ' '; i = i * 10 + *p++ - '0');
    mrph->bunrui = i; p++;
#endif
    /* ѷ No. */
    for (i = 0; *p != ' '; i = i * 10 + *p++ - '0');
    mrph->ktype = i; p++;
    /* ѷ No. */
    for (i = 0; *p != ' '; i = i * 10 + *p++ - '0');
    mrph->kform = i; p++;
    /* Ť */
#ifndef VGRAM
    if (*p < '0' || *p > '9')
      cha_exit(1, "The format of the patricia dictionary is invalid.");
#endif
    for (i = 0; *p != ' '; i = i * 10 + *p++ - '0');
    mrph->weight = i; p++;
    /* ɤ */
    s = strchr(p, ' ');
    *s = '\0';
    mrph->yomi = (char *)cha_malloc_char(s - p + 1);
    strcpy(mrph->yomi, p);
    p = s + 1;
    /* ³ơֹ֥ */
#ifdef VGRAM
    if (*p < '0' || *p > '9')
      cha_exit(1, "The format of the patricia dictionary is invalid.");
#endif
    for (i = 0; *p != ' '; i = i * 10 + *p++ - '0');
    mrph->con_tbl = i; p++;
    /* ̣ */
    s = strchr(p, '\n');
    *s = '\0';
    mrph->info = (char *)cha_malloc_char(s - p + 1);
    strcpy(mrph->info, p);

    ret = s + 1;

#elif 0
    ret = strchr(pbuf, '\n') + 1;
    /* Ф(ĹΤߤ) */
    mrph->base_length = mrph->length = strlen(strtok(pbuf, "\t"));
    /* ʻ No. */
    mrph->hinsi = atoi(strtok(NULL," "));
    /* ʻʬ No. */
    mrph->bunrui = atoi(strtok(NULL," "));
    /* ѷ No. */
    mrph->ktype = atoi(strtok(NULL," "));
    /* ѷ No. */
    mrph->kform = atoi(strtok(NULL," "));
    /* Ť */
    mrph->weight = atoi(strtok(NULL," "));
    /* ɤ */
    s = strtok(NULL, " ");
    mrph->yomi = (char *)cha_malloc_char(strlen(s) + 1);
    strcpy(mrph->yomi, s);
    /* ³ơֹ֥ */
    mrph->con_tbl = atoi(strtok(NULL," "));
    /* ̣ */
    s = strtok(NULL, "\n");
    mrph->info = (char *)cha_malloc_char(strlen(s) + 1);
    strcpy(mrph->info, s);
#else
    char midasi[256];
    char dummy[256];
    int hinsi, bunrui, ktype, kform;

    ret = strchr(pbuf, '\n') + 1;
    sscanf(pbuf, "%s %d %d %d %d %d %s %d %s",
	   midasi,

	   &hinsi, &bunrui, &ktype,
	   &kform, &mrph->weight, mrph->yomi,
	   &mrph->con_tbl, dummy);
    mrph->base_length = mrph->length = strlen(midasi);

    mrph->hinsi = hinsi;
    mrph->bunrui = bunrui;
    mrph->ktype = ktype;
    mrph->kform = kform;
#endif

    mrph->midasi = target;
    mrph->is_undef = 0;

    return ret;
}

/***********************************************************************
 * compare_top_str1 - compare strings to prefix of s1
 ***********************************************************************/
static int compare_top_str1(s1, s2)
    char *s1, *s2;
{
    /* 1996.07.28 by akira-k */
    for (;;) {
	if (*s1 == '\0') return TRUE;
	if (*s1++ != *s2++) return FALSE;
    }
}

/***********************************************************************
 * katuyou_process - find all forms which match the follow string
 ***********************************************************************/
static int *katuyou_process(follows, ktype)
    char *follows;
    int ktype;
{
    static int formstr[FORM_NO];
    int *fs = formstr;
    int f;

    for (f = 1; Form[ktype][f].name; f++)
      if (compare_top_str1(Form[ktype][f].gobi, follows))
	*fs++ = f;
    *fs = 0;

    return formstr;
}

/***********************************************************************
 * register_mrph - ѤĴ٤ʤ1ĤηǤХåեɲ
 *
 * mrph_nump: ХåեɲäǤο
 ************************************************************************/
static int register_mrph(mrph_idx, nmrphp)
    int mrph_idx, *nmrphp;
{
    int new_mrph_idx = mrph_idx;
    mrph2_t *new_mrph = &Mrph[mrph_idx];

    if (
#ifdef VGRAM
	Hinsi[new_mrph->hinsi].kt
#else
	Class[new_mrph->hinsi][new_mrph->bunrui].kt
#endif
	) {
	/* Ѥ */
	if (new_mrph->kform == 0) {
	    /* 촴 */
	    int *f;
	    int ktype = new_mrph->ktype;
	    int length = new_mrph->length;
	    int con_tbl = new_mrph->con_tbl;
	    char *follows = new_mrph->midasi + new_mrph->base_length;
	    mrph2_t *mrph0 = new_mrph;
	    for (f = katuyou_process(follows, ktype); *f; f++) {
		if (new_mrph != mrph0)
		  *new_mrph = *mrph0;
		new_mrph->kform = *f;
		new_mrph->length = length + strlen(Form[ktype][*f].gobi);
		new_mrph->con_tbl = con_tbl + *f - 1;
		new_mrph++;
		if (++new_mrph_idx == MRPH_NUM)
		  return FALSE;
	    }
	} else {
	    /* 촴ʤ */
	    new_mrph->base_length = 0;
	    new_mrph->yomi = "";
	    if (++new_mrph_idx == MRPH_NUM)
	      return FALSE;
	}
    } else {                         /* Ѥʤ */
	if (++new_mrph_idx == MRPH_NUM)
	  return FALSE;
    }

    *nmrphp = new_mrph_idx - mrph_idx;

    return TRUE;
}

/***********************************************************************
 * convert_mrphs - ѤĴ٤ʤǤХåեɲ
 *
 * mrph_nump: ХåեɲäǤο
 ***********************************************************************/
static int convert_mrphs(target, pat_buffer, mrph_idx, mrph_nump)
    char *target, *pat_buffer;
    int mrph_idx, *mrph_nump;
{
    int nmrph;
    int new_mrph_idx = mrph_idx;
    char *pbuf;

    for (pbuf = pat_buffer; *pbuf; ) {
	pbuf = take_data(&Mrph[new_mrph_idx], pbuf, target);
#if 0
	printf("%4d: ", new_mrph_idx);
	print_mrph(&Mrph[new_mrph_idx], 'f', NULL);
#endif
	if (register_mrph(new_mrph_idx, &nmrph) == FALSE)
	  return FALSE;
	new_mrph_idx += nmrph;
    }

    *mrph_nump = new_mrph_idx - mrph_idx;

    return TRUE;
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <trim_space> ڡκ (ȾѥڡΤߤѹ by T.U
                                                                   '96.01.11)
------------------------------------------------------------------------------
*/
static int trim_space(str, pos)
    char *str;
    int pos;
{
    for (;;) {
	/* Ⱦѥڡ */
	if (str[pos] == ' ')
	  pos++;
#if 0
	/* ѥڡ */
	else if ((unsigned char)str[pos  ] == 0xA1 &&
		 (unsigned char)str[pos+1] == 0xA1)
	  pos += 2;
#endif
	else
	  break;
    }
    return pos;
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <pos_match_process>
------------------------------------------------------------------------------
*/
static int pos_match_process(pos, p_idx)
    int pos, *p_idx;
{
    static int p_start;
    int i, j;

    j = 0;
    if (pos == 0) {
	/* new sentence */
	p_idx[j++] = 0;
	p_start = 1;
    } else {
	for (i = p_start; i < Path_num; i++) {
	    if (Path[i].end <= pos) {
		if (i == p_start)
		  p_start++;
		if (Path[i].end == pos)
		  p_idx[j++] = i;
	    }
	}
    }
    p_idx[j] = -1;

    return p_start;
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <check_connect>
------------------------------------------------------------------------------
*/
static int check_connect(pos, m_num, p_idx)
    int pos, m_num, *p_idx;
{
    typedef struct _path_cost_t {
	int min_cost;
	short min_cost_no;
	short state;
	short num;
	int   cost[PATH1_NUM];
	short pno[PATH1_NUM];
    } path_cost_t;

    static path_cost_t pcost[PATH1_NUM];
    int pcost_num;
    short path[PATH1_NUM], *new_path;
    mrph2_t *new_mrph;
    int i, pno, pcostno, npath;
    int	haba_cost, con_cost, cost, mrph_cost;
    int con_tbl, next_state;

#ifdef DEBUG
    printf("[m:%d] ", m_num);
#endif
    new_mrph = &Mrph[m_num];
    con_tbl = new_mrph->con_tbl;

    pcost[0].state = -1;
    pcost_num = 0;

    for (i = 0; (pno = p_idx[i]) >= 0; i++) {
	/* ȥޥȥĴ٤Ƽ֤³ȤФ */
	next_state = check_automaton
	  (Path[pno].state,
#ifdef KOCHA
	   Path[Path[pno].path[0]].state,
	   Mrph[Path[pno].mrph_p].midasi,
	   Mrph[Path[pno].mrph_p].is_undef,
#endif
	   con_tbl, Undef_con_cost, &con_cost);

#ifdef DEBUG
	printf("[m1:%d,m2:%d,cost:%d,state:%d]\n",
	       Path[pno].mrph_p,m_num,con_cost,next_state);
#endif
	if (con_cost == 0)
	  continue;

	/* cost ׻ */
	cost = Path[pno].cost + con_cost * Con_cost_weight;

	/* ɤ pcost °뤫Ĵ٤ */
	for (pcostno = 0; pcostno < pcost_num; pcostno++)
	  if (next_state == pcost[pcostno].state)
	    break;
	if (pcostno < pcost_num) {
	    if (pcost[pcostno].min_cost < cost - Cost_width)
	      continue;
	} else {
	    /*  pcost  */
	    pcost_num++;
	    pcost[pcostno].num = 0;
	    pcost[pcostno].state = next_state;
	    pcost[pcostno].min_cost = INT_MAX;
	}

	/* pcost Ͽ */
	pcost[pcostno].cost[pcost[pcostno].num] = cost;
	pcost[pcostno].pno[pcost[pcostno].num] = pno;
	if (cost < pcost[pcostno].min_cost) {
	    pcost[pcostno].min_cost = cost;
	    pcost[pcostno].min_cost_no = pcost[pcostno].num;
	}
	pcost[pcostno].num++;
    }

    if (pcost_num == 0)
      return TRUE;

    /* ǥ */
    if (new_mrph->is_undef)
      mrph_cost = Undef_cost /* *new_mrph->length */ ;
    else {
#ifdef VGRAM
	mrph_cost = Hinsi[new_mrph->hinsi].cost;
#else
	mrph_cost = Class[new_mrph->hinsi][new_mrph->bunrui].cost;
#endif
    }
    mrph_cost *= new_mrph->weight * Mrph_cost_weight;

#ifdef KOCHA
    if (new_mrph->midasi &&
	(unsigned char)new_mrph->midasi[0] == 0xA1 &&
	(unsigned char)new_mrph->midasi[1] == 0xA1) {
	for (i = 0; i < pcost[0].num; i++) {
	    new_path = cha_malloc_short(2);
	    new_path[0] = pcost[0].pno[i];
	    new_path[1] = -1;

	    Path[Path_num].cost = pcost[0].cost[i] + mrph_cost;
	    Path[Path_num].mrph_p = m_num;
	    Path[Path_num].state = pcost[0].state;
	    Path[Path_num].start = pos;
	    Path[Path_num].end = pos + new_mrph->length;
	    Path[Path_num].path = new_path;
#ifdef KOCHA_DEBUG
	    printf("[Sp:%d,prev:%d,m:%d,c:%d,pc:%d]\n",
		   Path_num,Path[Path_num].path[0],m_num,pcost[0].cost[i],Path[Path_num].cost);
#endif
	    if (++Path_num == PATH_NUM)
	      return FALSE;
	}
    } else {
#endif /* KOCHA */
	for (pcostno = 0; pcostno < pcost_num; pcostno++) {
	    /* ˤޤäƤѥȴФ */
	    haba_cost = pcost[pcostno].min_cost + Cost_width;
	    npath = 0;
	    path[npath++] = pcost[pcostno].pno[pcost[pcostno].min_cost_no];
	    for (i = 0; i < pcost[pcostno].num; i++)
	      if (pcost[pcostno].cost[i] <= haba_cost && i != pcost[pcostno].min_cost_no)
		path[npath++] = pcost[pcostno].pno[i];
	    path[npath++] = -1;

	    new_path = cha_malloc_short(npath);
	    memcpy(new_path, path, sizeof(short) * npath);

	    /* Path Ͽ */
	    Path[Path_num].cost = pcost[pcostno].min_cost + mrph_cost;
	    Path[Path_num].mrph_p = m_num;
	    Path[Path_num].state = pcost[pcostno].state;
	    Path[Path_num].start = pos;
	    Path[Path_num].end = pos + new_mrph->length;
	    Path[Path_num].path = new_path;
#ifdef DEBUG
	    printf("[p:%d,prev:%d,m:%d,c:%d,pc:%d]\n",
		   Path_num,Path[Path_num].path[0],m_num,pcost[0].cost[i],Path[Path_num].cost);
#endif
	    if (++Path_num == PATH_NUM)
	      return FALSE;
	}
#ifdef KOCHA
    }
#endif
    return TRUE;
}

/*
------------------------------------------------------------------------------
  PROCEDURE: <chasen_sent> ʸǲϤ    by T.Utsuro
------------------------------------------------------------------------------
*/
int chasen_sent(target)
    char *target;
{
    int  i, dic_no;
    int  pos, pos_end, length;
    int  path_idx[PATH_NUM];
    int  mrph_idx, new_mrph_idx;
    int  is_new_sent, is_2byte;
    int  undef_len, need_undef;
    char pat_buffer[50000];
    short path0 = -1;
#ifdef _WIN32
    char euc_target[CHA_INPUT_SIZE];
#endif

    /* ԡιԤϥå */
    /*	Ⱦѥڡ, by. T.U. '96.01.11 */
    target += trim_space(target, 0);
    if (!target[0])
      return FALSE;

#ifdef _WIN32
    strcpy(euc_target,target);
    sjis2euc(euc_target);
    hankana2zenkana(target);
#endif

    length = strlen(target);
    is_new_sent = 1; /* ʸ */

    cha_buffer_free();

    /* ʸƬ */
    Path[0].end = 0;
    Path[0].path = &path0;
    Path[0].cost = 0;
    Path[0].mrph_p = 0;
    Path[0].state = 0;
    Path_num = 1;

    Mrph[0].midasi = ""; /* for KOCHA */
    Mrph[0].hinsi = 0;
#ifndef VGRAM
    Mrph[0].bunrui = 0;
#endif
    Mrph[0].con_tbl = 0;
    Mrph[0].is_undef = 0;
    Mrph[0].weight = MRPH_DEFAULT_WEIGHT;

    new_mrph_idx = mrph_idx = 1;

    for (pos = 0; pos < length;
	 pos += is_2byte ? 2 : Mrph[mrph_idx-1].length) {
	/* Added by Kurohashi. */
	pos_end = pos;
	/*  Ⱦѥڡ, by. T.U. '96.01.11 */
	if ((pos = trim_space(target, pos_end)) == length) {
	    length = pos_end;
	    break;
	}

        is_2byte = (unsigned char)target[pos] & 0x80;

	pos_match_process(pos_end, path_idx);
	if (path_idx[0] < 0)
	  continue;

	/* (ʸΤ߸) */
	if (is_2byte) {
	    for(dic_no = 0; dic_no < number_of_tree; dic_no++) {
		int nmrph;
		/* ѥȥꥷڤǤ򸡺 */
		pat_buffer[0] = '\0';
#ifdef _WIN32
		pat_search(dic_file[dic_no], euc_target + pos,
			   &tree_top[dic_no], pat_buffer);
		euc2sjis((unsigned char *)pat_buffer);
#else
		pat_search(dic_file[dic_no], target + pos,
			   &tree_top[dic_no], pat_buffer);
#endif
		/* ѤĤķǤ Mrph ɲ */
		if (convert_mrphs(target + pos, pat_buffer, new_mrph_idx, &nmrph)
		    == FALSE)
		  goto error_end;
		new_mrph_idx += nmrph;
	    }
	}

#if 0
	/*
	 * ̤ϢܥȤ 0 ΤȤñ줬1Ĥ⼭Ǥʤä
	 * ̤ Mrph ɲ
	 */
	if (Undef_con_cost == 0 || mrph_idx == new_mrph_idx) {
	    if (register_undef_mrph(target + pos, new_mrph_idx) == FALSE)
	      goto error_end;
	    new_mrph_idx++;
	}
#endif

	/* ̤ĹĴ٤Ƥ */
#ifdef _WIN32
	undef_len = undef_mrph_len(euc_target + pos);
#else
	undef_len = undef_mrph_len(target + pos);
#endif
	/* ̤νɬפɤ */
	need_undef = 1;

	/* ľΥѥȤ³å */
	for (i = mrph_idx; i < new_mrph_idx; i++) {
	    if (Mrph[i].length == undef_len)
	      need_undef = 0;
	    if (check_connect(pos, i, path_idx) == FALSE)
	      goto error_end;
	}

	/* ̤ */
	if (Undef_con_cost == 0 || need_undef) {
	    if (register_undef_mrph(target + pos, new_mrph_idx, undef_len) == FALSE)
	      goto error_end;
	    if (check_connect(pos, new_mrph_idx, path_idx) == FALSE)
	      goto error_end;
	    new_mrph_idx++;
	}

	mrph_idx = new_mrph_idx;
    }

    /* ʸ */
    Mrph[mrph_idx].midasi = ""; /* for KOCHA */
    Mrph[mrph_idx].hinsi = 0;
#ifndef VGRAM
    Mrph[mrph_idx].bunrui = 0;
#endif
    Mrph[mrph_idx].con_tbl = 0;
    Mrph[mrph_idx].is_undef = 0;
    Mrph[mrph_idx].weight = MRPH_DEFAULT_WEIGHT;
    if (++mrph_idx == MRPH_NUM)
      goto error_end;

    pos_match_process(length, path_idx);

    if (check_connect(length, mrph_idx-1, path_idx) == FALSE)
      goto error_end;

#ifdef DEBUG
    for (i = 1; i < mrph_idx - 1; i++) {
	printf("%4d: %4d ", i, Mrph[i].con_tbl);
	print_mrph(&Mrph[i], 'F', "%-11m %-11y %-11P3- %-14T  %F \n");
    }
#endif

#ifdef DEBUG
    printf("num - mrph:%d path:%d, size - mrph:%d path:%d \n",
	   mrph_idx, Path_num, sizeof(mrph2_t), sizeof(path_t));
#endif

    return TRUE;

  error_end:
    fprintf(stderr, "Too many morphs <%s>\n", target);
    return FALSE;
}

